@@ -20,15 +20,19 @@
 static const int debug = 0;
 static const int quiet = 0;
 
+static EnchantBroker *broker = NULL;
+static int broker_ref_cnt;
+
 struct _GtkSpell {
 	GtkTextView *view;
+	GtkTextBuffer *buffer;
 	GtkTextTag *tag_highlight;
 	GtkTextMark *mark_insert_start;
 	GtkTextMark *mark_insert_end;
 	gboolean deferred_check;
-	EnchantBroker *broker;
 	EnchantDict *speller;
 	GtkTextMark *mark_click;
+	gchar *lang;
 };
 
 static void gtkspell_free(GtkSpell *spell);
@@ -269,12 +273,9 @@ static void
 add_to_dictionary(GtkWidget *menuitem, GtkSpell *spell) {
 	char *word;
 	GtkTextIter start, end;
-	GtkTextBuffer *buffer;
-	
-	buffer = gtk_text_view_get_buffer(spell->view);
 
-	get_word_extents_from_mark(buffer, &start, &end, spell->mark_click);
-	word = gtk_text_buffer_get_text(buffer, &start, &end, FALSE);
+	get_word_extents_from_mark(spell->buffer, &start, &end, spell->mark_click);
+	word = gtk_text_buffer_get_text(spell->buffer, &start, &end, FALSE);
 	
 	enchant_dict_add_to_pwl( spell->speller, word, strlen(word));
 
@@ -287,12 +288,9 @@ static void
 ignore_all(GtkWidget *menuitem, GtkSpell *spell) {
 	char *word;
 	GtkTextIter start, end;
-	GtkTextBuffer *buffer;
-	
-	buffer = gtk_text_view_get_buffer(spell->view);
 
-	get_word_extents_from_mark(buffer, &start, &end, spell->mark_click);
-	word = gtk_text_buffer_get_text(buffer, &start, &end, FALSE);
+	get_word_extents_from_mark(spell->buffer, &start, &end, spell->mark_click);
+	word = gtk_text_buffer_get_text(spell->buffer, &start, &end, FALSE);
 	
 	enchant_dict_add_to_session(spell->speller, word, strlen(word));
 
@@ -306,15 +304,12 @@ replace_word(GtkWidget *menuitem, GtkSpell *spell) {
 	char *oldword;
 	const char *newword;
 	GtkTextIter start, end;
-	GtkTextBuffer *buffer;
 	
 	if (!spell->speller)
 		return;
 
-	buffer = gtk_text_view_get_buffer(spell->view);
-
-	get_word_extents_from_mark(buffer, &start, &end, spell->mark_click);
-	oldword = gtk_text_buffer_get_text(buffer, &start, &end, FALSE);
+	get_word_extents_from_mark(spell->buffer, &start, &end, spell->mark_click);
+	oldword = gtk_text_buffer_get_text(spell->buffer, &start, &end, FALSE);
 	newword = gtk_label_get_text(GTK_LABEL(GTK_BIN(menuitem)->child));
 
 	if (debug) {
@@ -323,8 +318,10 @@ replace_word(GtkWidget *menuitem, GtkSpell *spell) {
 		g_print("\nnew word: '%s'\n", newword);
 	}
 
-	gtk_text_buffer_delete(buffer, &start, &end);
-	gtk_text_buffer_insert(buffer, &start, newword, -1);
+	gtk_text_buffer_begin_user_action(spell->buffer);
+	gtk_text_buffer_delete(spell->buffer, &start, &end);
+	gtk_text_buffer_insert(spell->buffer, &start, newword, -1);
+	gtk_text_buffer_end_user_action(spell->buffer);
 
 	enchant_dict_store_replacement(spell->speller, 
 			oldword, strlen(oldword),
@@ -333,22 +330,22 @@ replace_word(GtkWidget *menuitem, GtkSpell *spell) {
 	g_free(oldword);
 }
 
-static GtkWidget*
-build_suggestion_menu(GtkSpell *spell, GtkTextBuffer *buffer,
-                      const char *word) {
-	const char *suggestion;
-	GtkWidget *topmenu, *menu;
+/* This function populates suggestions at the top of the passed menu */
+static void
+add_suggestion_menus(GtkSpell *spell, GtkTextBuffer *buffer,
+                      const char *word, GtkWidget *topmenu) {
+	GtkWidget *menu;
 	GtkWidget *mi;
-	GtkWidget *hbox;
-	void *spelldata;
 	char **suggestions;
 	size_t n_suggs, i;
 	char *label;
 	
-	topmenu = menu = gtk_menu_new();
+	menu = topmenu;
 
 	if (!spell->speller)
-		return topmenu;
+		return;
+
+	gint menu_position = 0;
 
 	suggestions = enchant_dict_suggest(spell->speller, word, strlen(word), &n_suggs);
 
@@ -361,18 +358,16 @@ build_suggestion_menu(GtkSpell *spell, GtkTextBuffer *buffer,
 		mi = gtk_menu_item_new();
 		gtk_container_add(GTK_CONTAINER(mi), label);
 		gtk_widget_show_all(mi);
-		gtk_menu_shell_prepend(GTK_MENU_SHELL(menu), mi);
+		gtk_menu_shell_insert(GTK_MENU_SHELL(menu), mi, menu_position++);
 	} else {
 		/* build a set of menus with suggestions. */
+		gboolean inside_more_submenu = FALSE;
 		for (i = 0; i < n_suggs; i++ ) {
 			if (i > 0 && i % 10 == 0) {
-				mi = gtk_menu_item_new();
-				gtk_widget_show(mi);
-				gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
-
+				inside_more_submenu = TRUE;
 				mi = gtk_menu_item_new_with_label(_("More..."));
 				gtk_widget_show(mi);
-				gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
+				gtk_menu_shell_insert(GTK_MENU_SHELL(menu), mi, menu_position++);
 
 				menu = gtk_menu_new();
 				gtk_menu_item_set_submenu(GTK_MENU_ITEM(mi), menu);
@@ -381,16 +376,13 @@ build_suggestion_menu(GtkSpell *spell, GtkTextBuffer *buffer,
 			g_signal_connect(G_OBJECT(mi), "activate",
 					G_CALLBACK(replace_word), spell);
 			gtk_widget_show(mi);
-			gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
+			if (inside_more_submenu) gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
+			else gtk_menu_shell_insert(GTK_MENU_SHELL(menu), mi, menu_position++);
 		}
 	}
 
-	enchant_dict_free_suggestions(spell->speller, suggestions);
-
-	/* Separator */
-	mi = gtk_menu_item_new();
-	gtk_widget_show(mi);
-	gtk_menu_shell_append(GTK_MENU_SHELL(topmenu), mi);
+	if (suggestions)
+		enchant_dict_free_string_list(spell->speller, suggestions);
 
 	/* + Add to Dictionary */
 	label = g_strdup_printf(_("Add \"%s\" to Dictionary"), word);
@@ -401,7 +393,7 @@ build_suggestion_menu(GtkSpell *spell, GtkTextBuffer *buffer,
 	g_signal_connect(G_OBJECT(mi), "activate",
 			G_CALLBACK(add_to_dictionary), spell);
 	gtk_widget_show_all(mi);
-	gtk_menu_shell_append(GTK_MENU_SHELL(topmenu), mi);
+	gtk_menu_shell_insert(GTK_MENU_SHELL(topmenu), mi, menu_position++);
 
 	/* - Ignore All */
 	mi = gtk_image_menu_item_new_with_label(_("Ignore All"));
@@ -410,43 +402,112 @@ build_suggestion_menu(GtkSpell *spell, GtkTextBuffer *buffer,
 	g_signal_connect(G_OBJECT(mi), "activate",
 			G_CALLBACK(ignore_all), spell);
 	gtk_widget_show_all(mi);
-	gtk_menu_shell_append(GTK_MENU_SHELL(topmenu), mi);
+	gtk_menu_shell_insert(GTK_MENU_SHELL(topmenu), mi, menu_position++);
+}
+
+static GtkWidget*
+build_suggestion_menu(GtkSpell *spell, GtkTextBuffer *buffer,
+                      const char *word) {
+	GtkWidget *topmenu;
+	topmenu = gtk_menu_new();
+	add_suggestion_menus(spell, buffer, word, topmenu);
 
 	return topmenu;
 }
 
+static void
+language_change_callback(GtkCheckMenuItem *mi, GtkSpell* spell) {
+	if (gtk_check_menu_item_get_active(mi)) {
+		GError* error = NULL;
+		gchar *name;
+		g_object_get(G_OBJECT(mi), "name", &name, NULL);
+		gtkspell_set_language(spell, name, &error);
+		g_free(name);
+	}
+}
+
+struct _languages_cb_struct {GList *langs;};
+
+static void
+dict_describe_cb(const char * const lang_tag,
+		 const char * const provider_name,
+		 const char * const provider_desc,
+		 const char * const provider_file,
+		 void * user_data) {
+
+	struct _languages_cb_struct *languages_cb_struct = (struct _languages_cb_struct *)user_data;
+
+	languages_cb_struct->langs = g_list_insert_sorted(
+		languages_cb_struct->langs, g_strdup(lang_tag),
+		(GCompareFunc) strcmp);
+}
+
+static GtkWidget*
+build_languages_menu(GtkSpell *spell) {
+	GtkWidget *active_item = NULL, *menu = gtk_menu_new();
+	GList *langs;
+	GSList *menu_group = NULL;
+
+	struct _languages_cb_struct languages_cb_struct;
+	languages_cb_struct.langs = NULL;
+
+	enchant_broker_list_dicts(broker, dict_describe_cb, &languages_cb_struct);
+
+	langs = languages_cb_struct.langs;
+
+	for (; langs; langs = langs->next) {
+		gchar *lang_tag = langs->data;
+		GtkWidget* mi = gtk_radio_menu_item_new_with_label(menu_group, lang_tag);
+		menu_group = gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(mi));
+
+		g_object_set(G_OBJECT(mi), "name", lang_tag, NULL);
+		if (strcmp(spell->lang, lang_tag) == 0)
+			active_item = mi;
+		else
+			g_signal_connect(G_OBJECT(mi), "activate",
+				G_CALLBACK(language_change_callback), spell);
+		gtk_widget_show(mi);
+		gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
+
+		g_free(lang_tag);
+	}
+	if (active_item)
+		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(active_item), TRUE);
+
+	g_list_free(languages_cb_struct.langs);
+
+	return menu;
+}
+
 static void
 populate_popup(GtkTextView *textview, GtkMenu *menu, GtkSpell *spell) {
-	GtkWidget *img, *mi;
-	GtkTextBuffer *buffer = gtk_text_view_get_buffer(textview);
+	GtkWidget *mi;
 	GtkTextIter start, end;
 	char *word;
 
+	/* menu separator comes first. */
+	mi = gtk_separator_menu_item_new();
+	gtk_widget_show(mi);
+	gtk_menu_shell_prepend(GTK_MENU_SHELL(menu), mi);
+
+	/* on top: language selection */
+	mi = gtk_menu_item_new_with_label(_("Languages"));
+	gtk_menu_item_set_submenu(GTK_MENU_ITEM(mi), build_languages_menu(spell));
+	gtk_widget_show_all(mi);
+	gtk_menu_shell_prepend(GTK_MENU_SHELL(menu), mi);
+
 	/* we need to figure out if they picked a misspelled word. */
-	get_word_extents_from_mark(buffer, &start, &end, spell->mark_click);
+	get_word_extents_from_mark(spell->buffer, &start, &end, spell->mark_click);
 
 	/* if our highlight algorithm ever messes up, 
 	 * this isn't correct, either. */
 	if (!gtk_text_iter_has_tag(&start, spell->tag_highlight))
 		return; /* word wasn't misspelled. */
 
-	/* menu separator comes first. */
-	mi = gtk_menu_item_new();
-	gtk_widget_show(mi);
-	gtk_menu_shell_prepend(GTK_MENU_SHELL(menu), mi);
-
-	/* then, on top of it, the suggestions menu. */
-	img = gtk_image_new_from_stock(GTK_STOCK_SPELL_CHECK, GTK_ICON_SIZE_MENU);
-	mi = gtk_image_menu_item_new_with_label(_("Spelling Suggestions"));
-	gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
-
-	word = gtk_text_buffer_get_text(buffer, &start, &end, FALSE);
-	gtk_menu_item_set_submenu(GTK_MENU_ITEM(mi),
-			build_suggestion_menu(spell, buffer, word));
+	/* then, on top of it, the suggestions */
+	word = gtk_text_buffer_get_text(spell->buffer, &start, &end, FALSE);
+	add_suggestion_menus(spell, spell->buffer, word, GTK_WIDGET (menu) );
 	g_free(word);
-
-	gtk_widget_show_all(mi);
-	gtk_menu_shell_prepend(GTK_MENU_SHELL(menu), mi);
 }
 
 /* when the user right-clicks on a word, they want to check that word.
@@ -457,18 +518,17 @@ button_press_event(GtkTextView *view, GdkEventButton *event, GtkSpell *spell) {
 	if (event->button == 3) {
 		gint x, y;
 		GtkTextIter iter;
-		GtkTextBuffer *buffer = gtk_text_view_get_buffer(view);
 
 		/* handle deferred check if it exists */
 		if (spell->deferred_check)
-			check_deferred_range(spell, buffer, TRUE);
+			check_deferred_range(spell, spell->buffer, TRUE);
 
 		gtk_text_view_window_to_buffer_coords(view, 
 				GTK_TEXT_WINDOW_TEXT, 
 				event->x, event->y,
 				&x, &y);
 		gtk_text_view_get_iter_at_location(view, &iter, x, y);
-		gtk_text_buffer_move_mark(buffer, spell->mark_click, &iter);
+		gtk_text_buffer_move_mark(spell->buffer, spell->mark_click, &iter);
 	}
 	return FALSE; /* false: let gtk process this event, too.
 					 we don't want to eat any events. */
@@ -480,46 +540,56 @@ button_press_event(GtkTextView *view, GdkEventButton *event, GtkSpell *spell) {
 static gboolean
 popup_menu_event(GtkTextView *view, GtkSpell *spell) {
 	GtkTextIter iter;
-	GtkTextBuffer *buffer = gtk_text_view_get_buffer(view);
 
-	gtk_text_buffer_get_iter_at_mark(buffer, &iter, 
-			gtk_text_buffer_get_insert(buffer));
-	gtk_text_buffer_move_mark(buffer, spell->mark_click, &iter);
+	gtk_text_buffer_get_iter_at_mark(spell->buffer, &iter, 
+			gtk_text_buffer_get_insert(spell->buffer));
+	gtk_text_buffer_move_mark(spell->buffer, spell->mark_click, &iter);
 	return FALSE; /* false: let gtk process this event, too. */
 }
 
+static void
+_set_lang_from_dict(const char * const lang_tag,
+		    const char * const provider_name,
+		    const char * const provider_desc,
+		    const char * const provider_dll_file,
+		    void * user_data)
+{
+	GtkSpell *spell = user_data;
+
+	g_free(spell->lang);
+	spell->lang = g_strdup(lang_tag);
+}
+
 static gboolean
 gtkspell_set_language_internal(GtkSpell *spell, const gchar *lang, GError **error) {
+	EnchantDict *dict;
 
 	if (lang == NULL) {
 		lang = g_getenv("LANG");
 		if (lang) {
-			if (g_strncasecmp(lang, "C", 1) == 0)
+			if ((strcmp(lang, "C") == 0) || (strcmp(lang, "c") == 0))
 				lang = NULL;
 			else if (lang[0] == 0)
 				lang = NULL;
 		}
 	}
 
-	if (!spell->broker)
-		spell->broker = enchant_broker_init();
-
-	if (spell->speller) {
-		enchant_broker_free_dict(spell->broker, spell->speller);
-		spell->speller = NULL;
-	}
-
-	if (!lang) {
+	if (!lang)
 		lang = "en";
+
+	dict = enchant_broker_request_dict(broker, lang);
+
+	if (!dict) {
+		g_set_error(error, GTKSPELL_ERROR, GTKSPELL_ERROR_BACKEND,
+			_("enchant error for language: %s"), lang);
+		return FALSE;
 	}
 
-	spell->speller = enchant_broker_request_dict(spell->broker, lang );
+	if (spell->speller)
+		enchant_broker_free_dict(broker, spell->speller);
+	spell->speller = dict;
 
-	if (!spell->speller) {
-		g_set_error(error, GTKSPELL_ERROR, GTKSPELL_ERROR_BACKEND,
-			_("enchant error for language: %s"),lang);
-		return FALSE;
-	} 
+	enchant_dict_describe(dict, _set_lang_from_dict, spell);
 
 	return TRUE;
 }
@@ -527,7 +597,7 @@ gtkspell_set_language_internal(GtkSpell *spell, const gchar *lang, GError **erro
 /**
  * gtkspell_set_language:
  * @spell:  The #GtkSpell object.
- * @lang: The language to use, in a form pspell understands (it appears to
+ * @lang: The language to use, in a form enchant understands (it appears to
  * be a locale specifier?).
  * @error: Return location for error.
  *
@@ -558,14 +628,100 @@ gtkspell_set_language(GtkSpell *spell, const gchar *lang, GError **error) {
  */
 void
 gtkspell_recheck_all(GtkSpell *spell) {
-	GtkTextBuffer *buffer;
 	GtkTextIter start, end;
 
-	buffer = gtk_text_view_get_buffer(spell->view);
+	gtk_text_buffer_get_bounds(spell->buffer, &start, &end);
 
-	gtk_text_buffer_get_bounds(buffer, &start, &end);
+	check_range(spell, spell->buffer, start, end, TRUE);
+}
 
-	check_range(spell, buffer, start, end, TRUE);
+/* changes the buffer
+ * a NULL buffer is acceptable and will only release the current one */
+static void
+gtkspell_set_buffer(GtkSpell *spell, GtkTextBuffer *buffer)
+{
+	GtkTextTagTable *tagtable;
+	GtkTextIter start, end;
+
+	if (spell->buffer) {
+		g_signal_handlers_disconnect_matched(spell->buffer,
+				G_SIGNAL_MATCH_DATA,
+				0, 0, NULL, NULL,
+				spell);
+
+		tagtable = gtk_text_buffer_get_tag_table(spell->buffer);
+
+		gtk_text_buffer_get_bounds(spell->buffer, &start, &end);
+		gtk_text_buffer_remove_tag(spell->buffer, spell->tag_highlight, &start, &end);
+		spell->tag_highlight = NULL;
+
+		gtk_text_buffer_delete_mark(spell->buffer, spell->mark_insert_start);
+		spell->mark_insert_start = NULL;
+		gtk_text_buffer_delete_mark(spell->buffer, spell->mark_insert_end);
+		spell->mark_insert_end = NULL;
+		gtk_text_buffer_delete_mark(spell->buffer, spell->mark_click);
+		spell->mark_click = NULL;
+
+		g_object_unref (spell->buffer);
+	}
+
+	spell->buffer = buffer;
+
+	if (spell->buffer) {
+		g_object_ref (spell->buffer);
+
+		g_signal_connect(G_OBJECT(spell->buffer),
+				"insert-text",
+				G_CALLBACK(insert_text_before), spell);
+		g_signal_connect_after(G_OBJECT(spell->buffer),
+				"insert-text",
+				G_CALLBACK(insert_text_after), spell);
+		g_signal_connect_after(G_OBJECT(spell->buffer),
+				"delete-range",
+				G_CALLBACK(delete_range_after), spell);
+		g_signal_connect(G_OBJECT(spell->buffer),
+				"mark-set",
+				G_CALLBACK(mark_set), spell);
+
+		tagtable = gtk_text_buffer_get_tag_table(spell->buffer);
+		spell->tag_highlight = gtk_text_tag_table_lookup(tagtable, GTKSPELL_MISSPELLED_TAG);
+
+		if (spell->tag_highlight == NULL) {
+			spell->tag_highlight = gtk_text_buffer_create_tag(spell->buffer,
+					GTKSPELL_MISSPELLED_TAG,
+#ifdef HAVE_PANGO_UNDERLINE_ERROR
+					"underline", PANGO_UNDERLINE_ERROR,
+#else
+					"foreground", "red", 
+					"underline", PANGO_UNDERLINE_SINGLE,
+#endif
+					NULL);
+		}
+
+		/* we create the mark here, but we don't use it until text is
+		 * inserted, so we don't really care where iter points.  */
+		gtk_text_buffer_get_bounds(spell->buffer, &start, &end);
+		spell->mark_insert_start = gtk_text_buffer_create_mark(spell->buffer,
+				"gtkspell-insert-start",
+				&start, TRUE);
+		spell->mark_insert_end = gtk_text_buffer_create_mark(spell->buffer,
+				"gtkspell-insert-end",
+				&start, TRUE);
+		spell->mark_click = gtk_text_buffer_create_mark(spell->buffer,
+				"gtkspell-click",
+				&start, TRUE);
+			
+		spell->deferred_check = FALSE;
+
+		/* now check the entire text buffer. */
+		gtkspell_recheck_all(spell);
+	}
+}
+
+static void
+buffer_changed (GtkTextView *view, GParamSpec *pspec, GtkSpell *spell)
+{
+	gtkspell_set_buffer(spell, gtk_text_view_get_buffer(view));
 }
 
 /**
@@ -581,10 +737,6 @@ gtkspell_recheck_all(GtkSpell *spell) {
  */
 GtkSpell*
 gtkspell_new_attach(GtkTextView *view, const gchar *lang, GError **error) {
-	GtkTextBuffer *buffer;
-	GtkTextTagTable *tagtable;
-	GtkTextIter start, end;
-
 	GtkSpell *spell;
 
 #ifdef ENABLE_NLS
@@ -598,10 +750,25 @@ gtkspell_new_attach(GtkTextView *view, const gchar *lang, GError **error) {
 	spell = g_object_get_data(G_OBJECT(view), GTKSPELL_OBJECT_KEY);
 	g_assert(spell == NULL);
 
+	/* We don't need to worry about thread safety.
+	 * Stuff shouldn't be attaching to a GtkTextView from anything other
+	 * than the mainloop thread */
+	if (!broker) {
+		broker = enchant_broker_init();
+		broker_ref_cnt = 0;
+	}
+	broker_ref_cnt++;
+
+
 	/* attach to the widget */
 	spell = g_new0(GtkSpell, 1);
 	spell->view = view;
 	if (!gtkspell_set_language_internal(spell, lang, error)) {
+		broker_ref_cnt--;
+		if (broker_ref_cnt == 0) {
+			enchant_broker_free(broker);
+			broker = NULL;
+		}
 		g_free(spell);
 		return NULL;
 	}
@@ -615,88 +782,35 @@ gtkspell_new_attach(GtkTextView *view, const gchar *lang, GError **error) {
 			G_CALLBACK(populate_popup), spell);
 	g_signal_connect(G_OBJECT(view), "popup-menu",
 			G_CALLBACK(popup_menu_event), spell);
+	g_signal_connect(G_OBJECT(view), "notify::buffer",
+			G_CALLBACK(buffer_changed), spell);
 
-	buffer = gtk_text_view_get_buffer(view);
+	spell->buffer = NULL;
+	gtkspell_set_buffer(spell, gtk_text_view_get_buffer(view));
 
-	g_signal_connect(G_OBJECT(buffer),
-			"insert-text",
-			G_CALLBACK(insert_text_before), spell);
-	g_signal_connect_after(G_OBJECT(buffer),
-			"insert-text",
-			G_CALLBACK(insert_text_after), spell);
-	g_signal_connect_after(G_OBJECT(buffer),
-			"delete-range",
-			G_CALLBACK(delete_range_after), spell);
-	g_signal_connect(G_OBJECT(buffer),
-			"mark-set",
-			G_CALLBACK(mark_set), spell);
-
-	tagtable = gtk_text_buffer_get_tag_table(buffer);
-	spell->tag_highlight = gtk_text_tag_table_lookup(tagtable, GTKSPELL_MISSPELLED_TAG);
-
-	if (spell->tag_highlight == NULL) {
-		spell->tag_highlight = gtk_text_buffer_create_tag(buffer,
-				GTKSPELL_MISSPELLED_TAG,
-#ifdef HAVE_PANGO_UNDERLINE_ERROR
-				"underline", PANGO_UNDERLINE_ERROR,
-#else
-				"foreground", "red", 
-				"underline", PANGO_UNDERLINE_SINGLE,
-#endif
-				NULL);
-	}
-
-	/* we create the mark here, but we don't use it until text is
-	 * inserted, so we don't really care where iter points.  */
-	gtk_text_buffer_get_bounds(buffer, &start, &end);
-	spell->mark_insert_start = gtk_text_buffer_create_mark(buffer,
-			"gtkspell-insert-start",
-			&start, TRUE);
-	spell->mark_insert_end = gtk_text_buffer_create_mark(buffer,
-			"gtkspell-insert-end",
-			&start, TRUE);
-	spell->mark_click = gtk_text_buffer_create_mark(buffer,
-			"gtkspell-click",
-			&start, TRUE);
-		
-	spell->deferred_check = FALSE;
-
-	/* now check the entire text buffer. */
-	gtkspell_recheck_all(spell);
 	return spell;
 }
 
 static void
 gtkspell_free(GtkSpell *spell) {
-	GtkTextBuffer *buffer;
-	GtkTextTagTable *table;
-	GtkTextIter start, end;
 
-	buffer = gtk_text_view_get_buffer(spell->view);
-	table = gtk_text_buffer_get_tag_table(buffer);
+	gtkspell_set_buffer(spell, NULL);
 
-	gtk_text_buffer_get_bounds(buffer, &start, &end);
-	gtk_text_buffer_remove_tag(buffer, spell->tag_highlight, &start, &end);
-
-	gtk_text_buffer_delete_mark(buffer, spell->mark_insert_start);
-	gtk_text_buffer_delete_mark(buffer, spell->mark_insert_end);
-	gtk_text_buffer_delete_mark(buffer, spell->mark_click);
-
-
-	if (spell->broker) {
+	if (broker) {
 		if (spell->speller) {
-			enchant_broker_free_dict(spell->broker, spell->speller);
+			enchant_broker_free_dict(broker, spell->speller);
+		}
+		broker_ref_cnt--;
+		if (broker_ref_cnt == 0) {
+			enchant_broker_free(broker);
+			broker = NULL;
 		}
-		enchant_broker_free(spell->broker);
 	}
 	g_signal_handlers_disconnect_matched(spell->view,
 			G_SIGNAL_MATCH_DATA,
 			0, 0, NULL, NULL,
 			spell);
-	g_signal_handlers_disconnect_matched(buffer,
-			G_SIGNAL_MATCH_DATA,
-			0, 0, NULL, NULL,
-			spell);
+	g_free(spell->lang);
 	g_free(spell);
 }
 
@@ -730,4 +844,44 @@ gtkspell_detach(GtkSpell *spell) {
 	gtkspell_free(spell);
 }
 
+/**
+ * gtkspell_get_suggestions_menu:
+ * @iter: Textiter of position in buffer to be corrected if necessary.
+ *
+ * Retrieves a submenu of replacement spellings, or NULL if the word at @iter is
+ * not misspelt.
+ *
+ * Returns: the #GtkMenu widget, or %NULL if there is no need for a menu
+ */
+GtkWidget*
+gtkspell_get_suggestions_menu(GtkSpell *spell, GtkTextIter *iter) {
+	GtkWidget *submenu = NULL;
+	GtkTextIter start, end;
 
+	g_return_val_if_fail(spell != NULL, NULL);
+
+	/* avoid an empty submenu when enchant is not working properly */
+	if (!spell->speller)
+		return NULL;
+
+	start = *iter;
+	/* use the same lazy test, with same risk, as does the default menu arrangement */
+	if (gtk_text_iter_has_tag(&start, spell->tag_highlight)) {
+		/* word was mis-spelt */
+		gchar *badword;
+		/* in case a fix is requested, move the attention-point */
+		gtk_text_buffer_move_mark(spell->buffer, spell->mark_click, iter);
+		if (!gtk_text_iter_starts_word(&start))
+			gtk_text_iter_backward_word_start(&start);
+		end = start;
+		if (gtk_text_iter_inside_word(&end))
+			gtk_text_iter_forward_word_end(&end);
+		badword = gtk_text_buffer_get_text (spell->buffer, &start, &end, FALSE);
+
+		submenu = build_suggestion_menu (spell, spell->buffer, badword);
+		gtk_widget_show (submenu);
+
+		g_free (badword);
+	}
+	return submenu;
+}
